from datetime import datetime
import logging
import pandas as pd
import numpy as np
from typing import Literal

from scipy.stats import norm
from sklearn import base
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import r2_score

from sam.feature_engineering import BuildRollingFeatures


logging.basicConfig(level=logging.INFO)


def engineer_steps(channel: str, channels: list):
    """Prepare feature engineering steps:

    Parameters
    ----------
    channel: str
        Name of channel of interest
    channels : list
        List of all channels in the data

    Returns
    -------
    list
        List of feature engineering steps to perform
    """
    other_discharge_channels = [c for c in channels if (c != channel) & ("discharge" in c.lower())]
    precipitation_channels = [c for c in channels if "precip" in c.lower()]

    discharge_step = (
        "lag_features_discharge_channels",
        BuildRollingFeatures(
            rolling_type="lag",
            window_size=[0, 1, 2, 3, 4, 5, 8],
            lookback=0,
            keep_original=False,
        ),
        other_discharge_channels,
    )

    precipitation_step = (
        "lag_features_precipitation_channels",
        BuildRollingFeatures(
            rolling_type="lag",
            window_size=[0, 3],
            lookback=0,
            keep_original=False,
        ),
        precipitation_channels,
    )

    if len(precipitation_channels) > 0:
        return [discharge_step, precipitation_step]

    return [discharge_step]


class ValidationModel(base.BaseEstimator, base.RegressorMixin):
    """Convenience class for building and running a MLPRegression model for anomaly
    detection. For a given channel, we take the top features and use all combinations of
    leaving one feature out to train the models. This way we can ensure we find true
    anomalies in the target channel rather than anomalies depending on one of the features.
    We accept anomalies that are flagged by all (or most) of those models.

    Parameters
    ----------
    df : pd.DataFrame
        Input data with channels as column names
    model_type : str (default='lasso')
        Could also be 'mlp'.
    training_end_date : datetime | str | None (default = None)
        Marks the endpoint of the training data and start of the test data
    epochs : int (default = 2)
        Number of epochs for MLP
    n_features : int (default = 5)
        Number of features to include in models + 1. So if you want to have 3 features
        in each sub-model this value should be 4.
    n_consensus: int | Literal["all"], (default = 'all')
        By default all sub-model predictions have to flag outliers, but you can also specify
        an integer of the number of models desired for consenus.
    outlier_window: int (default=1)
        the window size in which at least `outlier_limit` should be outside of `outlier_min_q`
    outlier_limit: int (default=1)
        the minimum number of outliers within outlier_window to be outside of `outlier_min_q`
    use_precipitation_features: bool (default=False)
        Whether or not to use precipitation features in models. If True, we will use all
        precipitation features in all models with lags of [0, 3]
    learning_rate : float (default=0.001)
    """

    def __init__(
        self,
        df: pd.DataFrame,
        model_type: str = "lasso",
        training_end_date: str | datetime | None = None,
        epochs: int = 2,
        n_features: int = 5,
        n_consensus: int | Literal["all"] = "all",
        outlier_window: int = 3,
        outlier_limit: int = 3,
        use_precipitation_features: bool = False,
        learning_rate: float = 0.001,
    ):
        self.df = df.copy()
        self.model_type = model_type
        self.discharge_channels = [c for c in self.df.columns if "discharge" in c.lower()]
        self.training_end_date = self._get_training_end_date(training_end_date)
        self.epochs = epochs
        self.n_features = n_features
        self.n_consensus = n_consensus
        self.outlier_window = outlier_window
        self.outlier_limit = outlier_limit
        self.use_precipitation_features = use_precipitation_features
        self.learning_rate = learning_rate

    def _get_training_end_date(self, training_end_date: datetime | str | None = None):
        """We can either provide a dataframe with both training and test set, in
        which case the parameter `training_end_date` denotes the cut-off; or we can
        provide only a training set, in which case training_end_date should be None.
        Then this method returns the last date of the input data.

        Parameters
        ----------
        training_end_date: datetime | str | None (default=None)

        Returns
        -------
        str
            The end date of the training data (input data on initialization)
            if `training_end_date` is None, otherwise `training_end_date`.
        """
        if training_end_date is None:
            return self.df.index.max()
        return training_end_date

    def _train_test_split(self, channel: str, training_end_date: datetime | str):
        """Split data into train and test sets based on datetime cutoff.
        Everything before the cutoff is training data, everything after is
        test data.

        Parameters
        ----------
        channel: str
            Name of channel of interest (target variable)
        training_end_date: datetime | str

        Returns
        -------
        X_train : pd.DataFrame
        y_train : pd.Series
        X_test : pd.DataFrame
        y_test : pd.Series
        """
        X_train = self.df.loc[:training_end_date, :].reset_index(drop=True)
        y_train = self.df.loc[:training_end_date, channel].reset_index(drop=True)
        X_test = self.df.loc[training_end_date:, :].reset_index(drop=True)
        y_test = self.df.loc[training_end_date:, channel].reset_index(drop=True)
        self.n_samples_train = len(y_train)
        self.n_samples_test = len(y_test)

        return X_train, y_train, X_test, y_test

    def _get_feature_channels(self, channel: str):
        """Get MLP model for particular channel

        Parameters
        ----------
        channel: str
            Name of channel of interest

        Returns
        -------
        feature_channels: list
            List of strings denoting the channels to use for feature engineering when
            predicting `channel`
        """
        discharge = self.df.loc[: self.training_end_date, self.discharge_channels]

        correlations = discharge.corr("spearman").loc[channel]
        sort_ids = np.argsort(correlations.values)
        feature_channels = correlations.index[sort_ids][-1 - self.n_features : -1]

        return feature_channels

    def _get_pipeline(
        self,
        channel: str,
        feature_channels: list,
    ):
        """Prepare ML pipeline

        Parameters
        ----------
        channel: str
            Name of channel of interest
        feature_channels: list
            List of channel names to use for feature engineering
        """
        engineer = ColumnTransformer(engineer_steps(channel, feature_channels), remainder="drop")
        scaler = StandardScaler()
        pipe = Pipeline(
            [
                ("columns", engineer),
                ("scaler", scaler),
                ("imputer", SimpleImputer(missing_values=np.nan, strategy="mean")),
            ]
        )
        return pipe

    def _get_mlp_model(self, channel: str, feature_channels: list):
        """Get MLP model for particular channel

        Parameters
        ----------
        channel: str
            Name of channel of interest
        feature_channels: list
            List of channel names to use for feature engineering

        Returns
        -------
        estimator : sklearn Model
        """
        from sam.models import MLPTimeseriesRegressor

        self.quantiles = (
            norm.cdf(3),
            norm.cdf(2),
            norm.cdf(1),
            1 - norm.cdf(1),
            1 - norm.cdf(2),
            1 - norm.cdf(3),
        )
        self.predict_ahead = (0,)

        estimator = MLPTimeseriesRegressor(
            predict_ahead=self.predict_ahead,
            quantiles=self.quantiles,
            feature_engineer=self._get_pipeline(channel, feature_channels),
            epochs=self.epochs,
            verbose=0,
            average_type="median",
            lr=self.learning_rate,
        )

        return estimator

    def _get_lasso_model(self, channel: str, feature_channels: list):
        """Get MLP model for particular channel

        Parameters
        ----------
        channel: str
            Name of channel of interest
        feature_channels: list
            List of channel names to use for feature engineering

        Returns
        -------
        estimator : sklearn Model
        """
        from darrow_poc.models.linear_quantile_regressor import LinearQuantileRegressor

        regressor = LinearQuantileRegressor(
            quantiles=[
                norm.cdf(3),
                norm.cdf(2),
                norm.cdf(1),
                norm.cdf(0),
                1 - norm.cdf(1),
                1 - norm.cdf(2),
                1 - norm.cdf(3),
            ],
        )

        estimator = Pipeline(
            [
                ("preprocessor", self._get_pipeline(channel, feature_channels)),
                ("regressor", regressor),
            ]
        )

        return estimator

    def _get_model(self, channel: str, feature_channels: list):
        """Get MLP model for particular channel

        Parameters
        ----------
        channel: str
            Name of channel of interest
        feature_channels: list
            List of channel names to use for feature engineering

        Returns
        -------
        estimator : sklearn Model
        """
        if self.model_type == "mlp":
            return self._get_mlp_model(channel, feature_channels)
        elif self.model_type == "lasso":
            return self._get_lasso_model(channel, feature_channels)
        else:
            raise NotImplementedError(
                f"model_type {self.model_type} not implemented." "Choose `mlp` or `lasso` instead."
            )

    def fit_and_evaluate(self, target_channel: str):
        """Fit ML model for each channel and evaluate performance on test set.

        Parameters
        ----------
        target_channel: str
            Name of target channel

        Returns
        -------
        model : dict
            Model object for each channel
        num_obs : dict
            Numbers of observations in training and test data
        pred : dict
            Predictions for each channel for test set
        r2 : pd.DataFrame
            Contains r2 scores
        """
        model, num_obs, pred, r2 = {target_channel: {}}, {}, {target_channel: {}}, {target_channel: {}}

        # Get training and test data
        X_train, y_train, X_test, y_test = self._train_test_split(target_channel, self.training_end_date)

        num_train = np.sum(~y_train.isna())
        num_test = np.sum(~y_test.isna())
        num_obs[f"train_{target_channel}"] = num_train
        num_obs[f"test_{target_channel}"] = num_test

        # Determine which channels to use for feature engineering
        feature_channels = self._get_feature_channels(target_channel)

        for leave_out_feature in feature_channels:
            feature_channel_subset = [c for c in feature_channels if c != leave_out_feature]

            logging.info(
                f"\nTraining model for channel {target_channel} in time period from "
                f"{X_train.index[0]} to {self.training_end_date}."
                f"\n We use the following feature channels: {feature_channel_subset}"
            )

            # Fit model
            model[target_channel][leave_out_feature] = self._get_model(target_channel, feature_channel_subset)
            model[target_channel][leave_out_feature].fit(X_train, y_train)

            if hasattr(model[target_channel][leave_out_feature], "model_"):
                from tensorflow.keras.optimizers import Adam

                model[target_channel][leave_out_feature].model_.compile(  # HACK
                    optimizer=Adam(learning_rate=self.learning_rate),
                    loss=None,
                )

            # Evaluate
            pred[target_channel][leave_out_feature] = self._standardize_prediction_column_names(
                model[target_channel][leave_out_feature].predict(X_test)
            )

            if isinstance(pred[target_channel][leave_out_feature], pd.DataFrame):
                y_hat = pred[target_channel][leave_out_feature].loc[:, "predict_lead_0_mean"]
            else:
                y_hat = pred[target_channel][leave_out_feature]
            finite_selection = ~y_test.isna() & ~np.isnan(y_hat)
            if finite_selection.sum() > 0:
                r2[target_channel][leave_out_feature] = r2_score(y_test[finite_selection], y_hat[finite_selection])
            else:
                r2[target_channel][leave_out_feature] = np.nan

        self.model = model
        self.num_obs = num_obs
        self.pred = pred
        self.r2 = r2

        return model, num_obs, pred, r2

    def predict(
        self,
        X: pd.DataFrame,
        target_channel,
    ):
        """Predict for each target channel and left out feature channel.

        Returns
        -------
        pred : dict
            Predictions for each channel for test set
        r2 : pd.DataFrame
            Contains r2 scores
        """
        pred = {target_channel: {}}
        X_test = X.drop(columns=[target_channel]).reset_index(drop=True)

        feature_channels = self._get_feature_channels(target_channel)

        for leave_out_feature in feature_channels:
            pred[target_channel][leave_out_feature] = self._standardize_prediction_column_names(
                self.model[target_channel][leave_out_feature].predict(X_test)
            )

        return pred

    def _flatten_output(self, output: dict[str, dict], name: str) -> dict[str, str]:
        return {
            f"{name}_target_{outer_key}_missing_feature_{inner_key}": f"{inner_value}"
            for outer_key, inner_dict in output.items()
            for inner_key, inner_value in inner_dict.items()
        }

    def _standardize_prediction_column_names(self, y_hat):
        return y_hat.rename(
            columns={
                "predict_q_0.9986501019683699": "predict_lead_0_q_0.9986501019683699",
                "predict_q_0.9772498680518208": "predict_lead_0_q_0.9772498680518208",
                "predict_q_0.8413447460685429": "predict_lead_0_q_0.8413447460685429",
                "predict_q_0.15865525393145707": "predict_lead_0_q_0.15865525393145707",
                "predict_q_0.02275013194817921": "predict_lead_0_q_0.02275013194817921",
                "predict_q_0.0013498980316301035": "predict_lead_0_q_0.0013498980316301035",
                "predict_q_0.5": "predict_lead_0_mean",
            }
        )



from os import PathLike
from pathlib import Path

import dill as pickle
import numpy as np
import pandas as pd

from twinn_ml_interface.input_data import InputData
from twinn_ml_interface.interface import ModelInterfaceV4
from twinn_ml_interface.objectmodels import (
    Configuration,
    DataLabelConfigTemplate,
    DataLevel,
    MetaDataLogger,
    ModelCategory,
    PredictionType,
    Tag,
    TrainWindowSizePriority,
    UnitTagTemplate,
    UnitTag,
    Unit,
    WindowViability,
)

from .anomaly_detection import get_anomalies
from .validation_model import ValidationModel


class POCAnomaly:
    model_type_name: str = "pocanomaly"
    # Model category is based on the output of the model.
    model_category: ModelCategory = ModelCategory.ANOMALY
    # Features used to train the model. If not supplied, equal to get_data_config_template().
    base_features: dict[DataLevel, list[UnitTag]] | None = None
    # This is only needed when get_target_template returns UnitTagTemplate
    target: UnitTag | None = None

    def __init__(self, target: UnitTag):
        self.target = target

    @staticmethod
    def get_target_template() -> UnitTagTemplate | UnitTag:
        """Get the UnitTag that will be the target of the model.

        Returns:
            UnitTagTemplate | UnitTag: The unit tag of the model target,
            either as template or as literal.
        """
        return UnitTag(Unit("DARROW_POC_DISCHARGE_STAH", "PURE_DARROW_POC_DISCHARGE", True), Tag("MEASUREMENT"))

    @staticmethod
    def get_data_config_template() -> list[DataLabelConfigTemplate]:
        """The specification of data needed to train and predict with the model.

        NOTE:
        Using DataLabelConfigTemplate is more complicated, but allows refering to relative relationships
        between units (e.g. selecting all children of the target). Specifying UnitTags directly is simpler,
        but requires you to know the exact units necessary for the model.

        Result:
            list[DataLabelConfigTemplate]: The data needed to train and predict with the model,
                either as template.
        """
        return [
            DataLabelConfigTemplate(
                data_level=DataLevel.SENSOR,
                unit_tag_templates=[
                    UnitTag.from_string("DARROW_POC_DISCHARGE_ALTENBURG1:MEASUREMENT"),
                    UnitTag.from_string("DARROW_POC_DISCHARGE_ESCHWEILER:MEASUREMENT"),
                    UnitTag.from_string("DARROW_POC_DISCHARGE_HERZOGENRATH1:MEASUREMENT"),
                    UnitTag.from_string("DARROW_POC_DISCHARGE_JUELICH:MEASUREMENT"),
                    UnitTag.from_string("DARROW_POC_DISCHARGE_STAH:MEASUREMENT"),
                    UnitTag.from_string("DARROW_POC_EVAPORATION_EVAP:MEASUREMENT"),
                    UnitTag.from_string("DARROW_POC_PRECIPITATION_MIDDENROER:MEASUREMENT"),
                    UnitTag.from_string("DARROW_POC_PRECIPITATION_URFT:MEASUREMENT"),
                ],
            ),
        ]

    @staticmethod
    def get_result_template() -> UnitTagTemplate | UnitTag:
        """The tag to post the predictions/results on.

        Returns:
           UnitTagTemplate, UnitTag: The unit tag of the model's output, either as template or as literal.
        """
        return UnitTag(Unit("DARROW_POC_DISCHARGE_STAH", "PURE_DARROW_POC_DISCHARGE", True), Tag("FORECAST"))

    @staticmethod
    def get_train_window_finder_config_template() -> (
        tuple[list[DataLabelConfigTemplate], TrainWindowSizePriority] | None
    ):
        """The config for running the train window finder.

        Returns:
            list[DataLabelConfigTemplate] | None: a template for getting the tags needed to run
                the train window finder. Defaults to None, then no train window finder will be
                used.
        """
        return None

    @classmethod
    def initialize(cls, configuration: Configuration, logger: MetaDataLogger) -> ModelInterfaceV4:
        """Post init function to pass metadata logger and some config to the model.

        NOTE:
        This is used, because we cannot inherit an __init__() from the Protocol, and because
        passing configuration and logger to each method where they are needed would be a little
        tedious.

        Args:
            configuration (Configuration): an API-like object to retrieve configuration.
            logger (MetaDataLogger): A MetaDataLogger object to write logs to MLflow later.
        """
        model = cls(configuration.target_name)
        model.configuration = configuration
        model.logger = logger
        return model

    def preprocess(self, input_data: InputData) -> InputData:
        """Preprocess input data before training.

        Args:
            data (InputData): Input data.

        Returns:
            InputData: Preprocessed input data.

        """
        return input_data

    def validate_input_data(
        self,
        input_data: InputData,
    ) -> WindowViability:
        """Validate if input data is usable for training.

        Args:
            data (InputData): Training data.

        Returns:
            WindowViability: For each PredictionType you get
                bool: Whether the data can be used for training. Default always true.
                str: Additional information about the window.
        """
        return {PredictionType.ML: (True, None)}

    def train(self, input_data: InputData, **kwargs) -> tuple[float, object]:
        """Train a model.

        Args:
            input_data (InputData): Preprocessed and validated training data.

        Returns:
            float: Number between (-inf, inf) indicating the model performance
            object: Any other object that can be used for testing. This object will be ignored
                by the infrastructure
        """

        def prep_string_for_mlflow(s: str) -> str:
            return s.replace(":MEASUREMENT", "").replace("DARROW_POC_", "")

        train = pd.concat(input_data.values(), axis=1)
        validator = ValidationModel(
            train,
            model_type="lasso",
            n_features=5,
            use_precipitation_features=False,
            training_end_date="2010-01-04 00:00:00",
        )
        _, num_obs, _, r2_by_target = validator.fit_and_evaluate(str(self.target))
        r2_by_missing_sensor = validator._flatten_output(r2_by_target, "r2")
        r2_log = {prep_string_for_mlflow(k): prep_string_for_mlflow(v) for k, v in r2_by_missing_sensor.items()}
        self.logger.log_params(r2_log)  # This will be logged to mlflow
        self.logger.log_params({f"samples_{prep_string_for_mlflow(k)}": str(v) for k, v in num_obs.items()})

        self._model = validator
        return np.mean([float(x) for x in r2_by_missing_sensor.values()]), None

    def predict(self, input_data: InputData, **kwargs) -> tuple[list[pd.DataFrame], object]:
        """Run a prediction with a trained model.

        Args:
            input_data (InputData): Prediction data.

        Returns:
            list[pd.DataFrame]: List of dataframes with predictions
            object: Any other object that can be used for testing. This object will be ignored
                by the infrastructure
        """
        target_channel = str(self.target)
        model = self._model
        X = pd.concat(input_data.values(), axis=1)
        predictions = model.predict(X, target_channel)
        anomalies = get_anomalies(predictions, X, target_channel)

        return [pd.DataFrame({target_channel: anomalies})], None

    def dump(self, foldername: PathLike, filename: str) -> None:
        """
        Writes the following files:
        * filename.pkl
        * filename.h5
        to the folder given by foldername.

        Args:
            foldername (PathLike): configurable folder name
            filename (str): name of the file
        """
        with open(Path(foldername) / (filename + ".pkl"), "wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def load(
        foldername: PathLike, filename: str, configuration: Configuration, logger: MetaDataLogger
    ) -> ModelInterfaceV4:
        """
        Reads the following files:
        * prefix.pkl
        * prefix.h5
        from the folder given by foldername.
        Output is an entire instance of the fitted model that was saved.
        Just as with `initialize`, configuration and logger are passed for you to use.

        Args:
            foldername (PathLike): configurable folder name
            filename (str): name of the file
            configuration (Configuration): an API-like object to retrieve configuration.
            logger (MetaDataLogger): A MetaDataLogger object to write logs to MLflow later.

        Returns:
            Model class with everything (except data) contained within to call the
            `predict()` method
        """
        with open(Path(foldername) / (filename + ".pkl"), "rb") as f:
            model = pickle.load(f)

        model.configuration = configuration
        model.logger = logger

        return model



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, RegressorMixin

# Keep package independent of statsmodels
try:
    from statsmodels.regression.quantile_regression import QuantReg
except ImportError:
    pass


class LinearQuantileRegressor(BaseEstimator, RegressorMixin):
    """
    scikit-learn style wrapper for QuantReg
    Fits a linear quantile regression model, base idea from
    https://github.com/Marco-Santoni/skquantreg/blob/master/skquantreg/quantreg.py
    This class requires statsmodels

    Parameters
    ----------
    quantiles: list or float, default=[0.05, 0.95]
        Quantiles to fit, with `` 0 < q < 1 `` for each q in quantiles.
    tol: float, default=1e-3
        The tolerance for the optimization. The optimization stops
        when duality gap is smaller than the tolerance
    max_iter: int, default=1000
        The maximum number of iterations
    fit_intercept: bool, default=True
        Whether to calculate the intercept for this model. If set to false,
        no intercept will be used in calculations (e.g. data is expected to be
        already centered).

    Attributes
    ----------
    model_: statsmodel model
        The underlying statsmodel class
    model_result_: statsmodel results
        The underlying statsmodel results

    Examples
    --------
    >>> # Prepare data
    >>> data = read_knmi('2018-02-01', '2019-10-01', freq='hourly',
    >>>                 variables=['FH', 'FF', 'FX', 'T']).set_index('TIME')
    >>> y = data['T']
    >>> X = data.drop('T', axis=1)
    >>> # Fit model
    >>> model = LinearQuantileRegression()
    >>> model.fit(X, y)
    """

    def __init__(
        self,
        quantiles: list = [0.05, 0.95],
        tol: float = 1e-3,
        max_iter: int = 1000,
        fit_intercept: bool = True,
    ):
        self.quantiles = quantiles
        self.tol = tol
        self.max_iter = max_iter
        self.fit_intercept = fit_intercept

    def _fit_single_model(self, X, y, q):
        # Statsmodels requires to add constant columns manually
        # otherwise the models will not fit an intercept
        X = pd.DataFrame(X).copy()
        if self.fit_intercept:
            X = X.assign(const=1)  # f(x) = a + bX = a*const + b*X
        model_ = QuantReg(y, X)
        model_result_ = model_.fit(q, p_tol=self.tol, max_iter=self.max_iter)
        coef = model_result_.params
        pvalues = model_result_.pvalues
        return coef, pvalues

    def fit(self, X: np.array, y: np.array):
        """
        Fit a Linear Quantile Regression using statsmodels
        """
        if type(self.quantiles) is float:
            self.q_ = [self.quantiles]
        elif type(self.quantiles) is list:
            self.q_ = self.quantiles
        else:
            raise TypeError(f"Invalid type, quantile {self.quantiles} " f"should be a float or list of floats")
        self.prediction_cols = [f"predict_q_{q}" for q in self.quantiles]
        models_ = [self._fit_single_model(X, y, q) for q in self.quantiles]
        self.coef_ = [m[0] for m in models_]
        self.pvalue_ = [m[1] for m in models_]
        return self

    def predict(self, X: np.array):
        """
        Predict / estimate quantiles
        """
        preds = [pd.DataFrame(X).assign(const=1).multiply(c).sum(axis=1) for c in self.coef_]
        preds_df = pd.concat(preds, axis=1)
        preds_df.columns = self.prediction_cols
        return preds_df

    def score(self, X: np.array, y: np.array):
        """
        Default score function. Returns the tilted loss
        """
        y_pred = self.predict(X)
        scores = [tilted_loss(y_true=y, y_pred=y_pred[f"predict_q_{q}"], quantile=q) for q in self.quantiles]
        score = np.mean(scores)
        return score


def tilted_loss(y_true: np.array, y_pred: np.array, quantile: float = 0.5):
    """
    Calculate tilted, or quantile loss with numpy. Given a quantile q, and an error e,
    then tilted loss is defined as `(1-q) * |e|` if `e < 0`, and `q * |e|` if `e > 0`.

    This function is the same as the mean absolute error if q=0.5, which approximates the median.
    For a given value of q, the function that minimizes the tilted loss will be the q'th quantile
    function.

    Parameters
    ----------
    y_true: array-like of shape = (n_samples)
        True labels.
    y_pred: array-like of shape = (n_samples)
        Predictions. Must be same shape as `y_true`
    quantile: float, optional (default=0.5)
        The quantile to use when computing tilted loss.
        Must be between 0 and 1 for tilted loss to be positive.

    Returns
    -------
    float:
        The quantile loss

    Examples
    --------
    >>> import numpy as np
    >>> from sam.metrics import tilted_loss
    >>> actual = np.array([1, 2, 3, 4])
    >>> pred = np.array([0.9, 2.1, 2.9, 3.1])
    >>> tilted_loss(actual, pred, quantile=0.5)
    """
    y_true, y_pred = np.array(y_true), np.array(y_pred)
    e = y_true - y_pred
    return np.mean(np.maximum(quantile * e, (quantile - 1) * e), axis=-1)



from twinn_ml_interface.interface import ModelInterfaceV4
from twinn_ml_interface.objectmodels import MetaDataLogger

from darrow_poc.models.poc import POCAnomaly
from twinn_ml_interface.mocks import ConfigurationMock
import unittest

import logging


logging.basicConfig()
logging.getLogger().setLevel(logging.DEBUG)


class TestModelFollowsInterface(unittest.TestCase):
    def test_model_follows_interface(self):
        pocanomaly = POCAnomaly.initialize(ConfigurationMock("", "", {}, [], []), MetaDataLogger())
        if not isinstance(pocanomaly, ModelInterfaceV4):
            raise TypeError(f"{POCAnomaly.__name__} does not conform to the Protocol {ModelInterfaceV4.__name__}.")


if __name__ == "__main__":
    unittest.main()



from typing import Literal
import pandas as pd
import numpy as np


def get_outliers(
    y_true,
    y_hat,
    outlier_min_q: int = 3,
    outlier_window: int = 1,
    outlier_limit: int = 1,
):
    """Determine outliers, similar to `sam_quantile_plot` implementation of SAM. Methods referenced
    in the parameter descriptionsn also refer to SAM code.

    See: https://github.com/RoyalHaskoningDHV/sam

    Parameters
    ----------
    y_true: pd.Series
        Pandas Series containing the actual values. Should have same index as y_hat.
    y_hat: pd.DataFrame
        Dataframe returned by the MLPTimeseriesRegressor .predict() function.
        Columns should contain at least `predict_lead_x_mean`, where x is predict ahead
        and for each quantile: `predict_lead_x_q_y` where x is the predict_ahead, and
        y is the quantile. So e.g.:
        `['predict_lead_0_q_0.25, predict_lead_0_q_0.75, predict_lead_mean']`
    outlier_window: int (default=1)
        the window size in which at least `outlier_limit` should be outside of `outlier_min_q`
    outlier_limit: int (default=1)
        the minimum number of outliers within outlier_window to be outside of `outlier_min_q`

    Returns
    -------
    outliers : np.ndarray
        Array with true false values denoting outliers with true
    """
    if isinstance(y_true, pd.core.series.Series):
        y_true = y_true.values

    predict_ahead = 0
    these_cols = [c for c in y_hat.columns if "predict_lead_%d_q_" % predict_ahead in c]
    col_order = np.argsort([float(c.split("_")[-1]) for c in these_cols])
    n_quants = int((len(these_cols)) / 2)

    valid_low = y_hat[these_cols[col_order[n_quants - 1 - (outlier_min_q - 1)]]]
    valid_high = y_hat[these_cols[col_order[n_quants + (outlier_min_q - 1)]]]
    outliers = (y_true > valid_high) | (y_true < valid_low)
    outliers = outliers.astype(int)
    k = np.ones(outlier_window)
    outliers = (np.convolve(outliers, k, mode="full")[: len(outliers)] >= outlier_limit).astype(bool)

    return outliers


def get_outlier_consensus(
    y_true: pd.Series,
    pred: pd.DataFrame,
    target_channel: str,
    n_consensus: int | Literal["all"],
    outlier_min_q: int = 3,
    outlier_window: int = 1,
    outlier_limit: int = 1,
):
    """Determine outliers. We only consider values to be outliers when they occur in
    all or most sub-model predictions. For instance, we might fit 4 models for the target channel,
    where each time we leave one feature out. Then we consider those values outliers that
    are flagged by all or most sub-models.

    Parameters
    ----------
    y_true: pd.Series
        Pandas Series containing the actual values. Should have same index as y_hat.
    y_hat: pd.DataFrame
        Dataframe returned by the MLPTimeseriesRegressor .predict() function.
        Columns should contain at least `predict_lead_x_mean`, where x is predict ahead
        and for each quantile: `predict_lead_x_q_y` where x is the predict_ahead, and
        y is the quantile. So e.g.:
        `['predict_lead_0_q_0.25, predict_lead_0_q_0.75, predict_lead_mean']`
    target_channel: str
        Name of target channel to make predictions for
    n_consensus: int | Literal["all"], (default = 'all')
        By default all sub-model predictions have to flag outliers, but you can also specify
        an integer of the number of models desired for consenus.
    outlier_window: int (default=1)
        the window size in which at least `outlier_limit` should be outside of `outlier_min_q`
    outlier_limit: int (default=1)
        the minimum number of outliers within outlier_window to be outside of `outlier_min_q`

    Returns
    -------
    outliers : np.ndarray
        Array with true false values denoting outliers with true
    """
    outliers = []
    for left_out_channel in pred[target_channel].keys():
        y_hat = pred[target_channel][left_out_channel]
        outliers.append(
            get_outliers(
                y_true,
                y_hat,
                outlier_min_q=outlier_min_q,
                outlier_window=outlier_window,
                outlier_limit=outlier_limit,
            )
        )

    if n_consensus == "all":
        return np.array(outliers).all(axis=0)
    return np.array(outliers).sum(axis=0) >= n_consensus


def get_anomalies(
    pred: dict,
    df_test: pd.DataFrame,
    target_channel: str,
    n_consensus: int | Literal["all"] = "all",
    outlier_window: int = 3,
    outlier_limit: int = 3,
):
    """Get outliers for all features in df_test

    Parameters
    ----------
    pred : dict
        The `pred` key in the output dictionary from the `predict` method of the
        `ValidationModel` class.
        It contains predictions for each target channel for sub-models with single
        channels left out (pred[<target_channel>][<left_out_channel>])
    df_test : pd.DataFrame
        Data where to find anomalies
    target_channel: str
    n_consensus: int | Literal["all"], (default = 'all')
        By default all sub-model predictions have to flag outliers, but you can also specify
        an integer of the number of models desired for consenus.
    outlier_window: int (default=1)
        the window size in which at least `outlier_limit` should be outside of `outlier_min_q`
    outlier_limit: int (default=1)
        the minimum number of outliers within outlier_window to be outside of `outlier_min_q`

    Returns
    -------
    anomalies : np.ndarray
        Array with true false values denoting outliers with true
    """
    y_true = df_test.loc[:, target_channel]
    anomalies = get_outlier_consensus(
        y_true,
        pred,
        target_channel,
        n_consensus=n_consensus,
        outlier_window=outlier_window,
        outlier_limit=outlier_limit,
    )

    return anomalies



import logging
from pathlib import Path

from darrow_poc.models.poc import POCAnomaly
from twinn_ml_interface.mocks import LocalConfig, ExecutorMock, ConfigurationMock

import unittest


logging.basicConfig(level=logging.INFO)
logging.getLogger("sam").setLevel(logging.WARNING)


BASE_DIR = Path(__file__).parent.parent


class TestModelWithLocalExecutor(unittest.TestCase):
    def test_model_with_local_executor(self):
        config = LocalConfig(
            POCAnomaly,
            BASE_DIR / "tests/testing_data/train.parquet",
            BASE_DIR / "tests/testing_data/test.parquet",
            BASE_DIR / "output/models",
            "poc_model",
            BASE_DIR / "output/predictions/predictions.parquet",
        )
        infra_config = ConfigurationMock("stah:discharge", "", {}, [], [])
        executor = ExecutorMock(config, infra_config)
        executor.run_full_flow()


if __name__ == "__main__":
    unittest.main()
